@@ -1,10 +1,9 @@
-from __future__ import print_function, absolute_import
 import numpy as np
 from numba import cuda
-from numba.cuda.testing import unittest
+from numba.cuda.testing import unittest, CUDATestCase
 
 
-class TestCudaComplex(unittest.TestCase):
+class TestCudaComplex(CUDATestCase):
     def test_cuda_complex_arg(self):
         @cuda.jit('void(complex128[:], complex128)')
         def foo(a, b):
@@ -19,6 +18,6 @@ class TestCudaComplex(unittest.TestCase):
 
 
 if __name__ == '__main__':
-    main()
+    unittest.main()
 
 
